#!/usr/bin/env python

"""
Extract information from Python module. Write to JSON.

usage: extract_module_info.py [-h] [-f] [-v] [-p] module [outputfile]

positional arguments:
  module             The module from which to extract info.
  outputfile         The name of the file to write to. Default: sys.stdout.

optional arguments:
  -h, --help         show this help message and exit
  -f, --force        Overwrite output file without asking.
  -v, --verbose      Verbose -- Display additional info.
  -p, --prettyprint  Pretty print the JSON output.
"""

from __future__ import print_function
import sys
import os
import inspect
import argparse
import importlib
import json


def getdoc(obj):
    doc = ''
    doc = inspect.getdoc(obj)
    if not doc:
        doc = inspect.getcomments(obj)
    return doc


def save_method(name, callable):
    data = {
        'name': name,
        'doc': getdoc(callable),
        'sig': str(inspect.getargspec(callable)),
    }
    return data


def extract(module):
    info = {}
    members = inspect.getmembers(module)
    # extract the items and docs
    imports = [item for item in members if inspect.ismodule(item[1])]
    classes = [item for item in members if inspect.isclass(item[1])]
    functions = [item for item in members if inspect.isfunction(item[1])]
    info['module'] = {
        'name': module.__name__,
        'doc': getdoc(module),
    }
    info['imports'] = []
    info['classes'] = []
    info['functions'] = []
    for name, module_ in imports:
        try:
            import_path = inspect.getfile(module_)
        except TypeError:
            import_path = None
        import_info = {
            'name': name,
            'path': import_path
        }
        info['imports'].append(import_info)
    class_info_collection = info['classes']
    for name, class_ in classes:
        class_info = {
            'name': name,
            'doc': getdoc(class_),
            'methods': [],
            'static_methods': [],
        }
        methods = [
            item for item in inspect.getmembers(class_) if
            inspect.isfunction(item[1]) or inspect.ismethod(item[1])]
        class_info['methods'] = []
        for name, method_ in methods:
            method_info = save_method(name, method_)
            class_info['methods'].append(method_info)
        class_info_collection.append(class_info)
    function_info_collection = info['functions']
    for name, function_ in functions:
        function_info = save_method(name, function_)
        function_info_collection.append(function_info)
    return info


def extract_and_write(options):
    target = options.module
    mod_name, has_attrs, attrs = target.partition(":")
    module = importlib.import_module(mod_name)
    try:
        pass
    except Exception as exc:
        msg = "Failed to import {} ({}: {})".format(
            mod_name,
            type(exc).__name__,
            exc)
        print(msg, file=sys.stderr)
        exit(2)
    if module.__name__ in sys.builtin_module_names:
        print("Can't get info for builtin modules.", file=sys.stderr)
        exit(1)
    info = extract(module)
    if options.prettyprint:
        indent = 4
    else:
        indent = None
    json_info = json.dumps(info, indent=indent)
    if options.outputfile:
        outfile = makefile(options)
        outfile.write(json_info)
        outfile.write('\n')
        outfile.close()
    else:
        sys.stdout.write(json_info)
        sys.stdout.write('\n')


def makefile(options):
    if os.path.exists(options.outputfile) and not options.force:
        sys.exit('Output file exists.  Use "-f" to overwrite.')
    outfile = open(options.outputfile, 'w')
    return outfile


def make_arg_parser():
    desc = 'Extract information from Python module.  Write to JSON.'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        'module',
        help='The module from which to extract info.')
    parser.add_argument(
        'outputfile',
        nargs='?', default='',
        help='The name of the file to write to.  Default: sys.stdout.')
    parser.add_argument(
        '-f', '--force', action='store_true',
        help='Overwrite output file without asking.')
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Verbose -- Display additional info.')
    parser.add_argument(
        '-p', '--prettyprint', action='store_true',
        help='Pretty print the JSON output.')
    return parser


def main():
    parser = make_arg_parser()
    options = parser.parse_args()
    extract_and_write(options)


if __name__ == '__main__':
    #import ipdb; ipdb.set_trace()
    main()
